@@ -2,9 +2,10 @@ from .phases import canonicalize_phases, get_phase_restrictions
 from .structurefactors import (to_structurefactor,
                                from_structurefactor,
                                compute_structurefactor_multiplicity,
-                               is_centric)
+                               is_centric,
+                               is_absent)
 from .symop import apply_to_hkl, phase_shift
 from .rfree import add_rfree, copy_rfree
-from .asu import hkl_is_absent, hkl_to_asu, hkl_to_observed, in_asu
+from .asu import hkl_to_asu, hkl_to_observed, in_asu
 from .cell import compute_dHKL
 from .binning import bin_by_percentile
