@@ -140,7 +140,7 @@ def test_from_gemmi(data_gemmi):
     assert_frame_equal(result, expected)
     assert result.spacegroup == expected.spacegroup
     assert result.cell == expected.cell
-    assert result._cache_index_dtypes == expected._cache_index_dtypes
+    assert result._index_dtypes == expected._index_dtypes
         
 
 def test_get_phase_keys(data_fmodel):
